//! # config Create
//!
//! `config` is analyze command line and make options.
//! And, provide interface for other functions.
//!
//! このCrateはコマンドラインを解析して各種オプションを設定します。
//! 又、設定されたオプションを各関数が必要となる形で提供します。

use regex::Regex;

// default Screamer Radio Data file name
//
// Screamer Radioのデフォルトのデータファイル名
pub(crate) const DEFAULT_STATIONS_FILE: &str = "stations.json";

// default Radio Sure! Stations Data file name
//
// Radio Sure!のデフォルトのデータファイル名
pub(crate) const DEFAULT_RSD_FILE: &str = "stations-2022-03-07.rsd";

// default Screamer Radio user data file name
//
// Screamer Radioのデフォルトのユーザーデータファイル名
pub(crate) const DEFAULT_SCREAMERSTATIONS_FILE: &str = "ScreamerStations.json";

// Filter type for genre/nation
//   Nothing : no filtering
//   Include : use matching data
//   Exclude : use no-matching data
//
// ジャンルや国名のフィルタータイプ
//   Nothing : フィルタリングしない
//   Include : マッチした物を使用
//   Exclude : マッチしない物を使用
#[derive(PartialEq)]
enum FilterType {
    Nothing,
    Include,
    Exclude,
}

// filter for matching
// provide function of matching
//   filter_type is one of FilterType
//   re is what after Regex::new(pattern).unwrap()
//
// マッチング用フィルター
// 後述の関数群を使用して対象文字列がマッチするかどうかを判定する機能を提供
//   filter_typeはフィルタリングの種類
//   reはRegex::new(pattern).unwrap()済みの物
struct REFilter {
    filter_type: FilterType,
    re: Regex,
}

impl REFilter {
    // make REFilter with FilterType and pattern string.
    // pattern string is formatted as "patter1,pattern2,..."
    //  ex. "Jazz,Adult Contemporary,Latin"
    // if success, return Ok(REFilter). failure, return Err(err_msg_string)
    //
    // 与えられたFilterTypeとパターンを元にREFilterを作成
    // なおpatternは"pattern1,pattern2,.."の用に , で区切る
    //  例. "Jazz,Adult Contemporary,Latin"
    // 作成に成功した場合はOk(REFilter)を返す。
    // 失敗した場合はErr(err_msg_string)を返す。
    pub fn new(filter_type: FilterType, pattern: &str) -> Result<REFilter, String> {
        let re = if filter_type == FilterType::Nothing {
            Regex::new(".").unwrap() // dummy
        } else {
            // include/exculdeの場合、patternは必須
            if pattern.is_empty() {
                return Err("empty string not allowed".to_string());
            };
            // まずは , で分割
            let base_strs: Vec<&str> = pattern.split(',').collect();
            // 次はescape
            let mut re_strs: Vec<String> = Vec::new();
            for base_str in base_strs {
                re_strs.push(regex::escape(base_str));
            }
            // それを | で結合した物を使用して正規表現作成
            match Regex::new(&re_strs.join("|")) {
                Ok(re) => re,
                Err(_) => return Err(format!("pattern is bad : {}", pattern)),
            }
        };

        Ok(REFilter { filter_type, re })
    }

    // return bool for given parameter 'trg'.
    // if FilterType is Nothing, always return true
    // if FilterType is Include, return true when 'trg' is matchsing
    // if FilterType is Exclude, return true when 'trg' is NOT matchsing
    //
    // 与えられたtrgが対象かどうかをboolで返却する。
    // FilterTypeがNothingならば常にtrueを返す。
    // FilterTypeがIncludeならばmatchingした場合にtrueを返す。
    // FilterTypeがExcludeならばmatching「しない」場合にtrueを返す。
    pub fn is_match(&self, trg: &str) -> bool {
        match self.filter_type {
            FilterType::Nothing => true,
            FilterType::Include => self.re.is_match(trg),
            FilterType::Exclude => !self.re.is_match(trg),
        }
    }
}

/// Config is made from args, and provide parameters and functions.
///
/// 引数を元にconfigを作成して、必要なパラメーターや関数を提供します。
pub struct Config {
    stations_file: String,
    rsd_file: String,
    screamer_file: String,
    genre_filter: REFilter,
    nation_filter: REFilter,
}

impl Config {
    /// With Strings Vector(these are env::args), setup needed parameters.
    /// if args include -h or -hj or -v, return error
    ///
    /// 文字列ベクター(実体は引数)を元に必要とされるパラメーターを設定します。
    /// 引数に -h か -hj か -v が有る場合はエラーを返します。
    ///
    /// # Examples
    /// ```
    /// use std::env;
    /// use std::process;
    /// use rs2sr::config::*;
    ///
    /// let args:Vec<String> = env::args().collect();
    ///
    /// let cfg = match Config::new(args) {
    ///     Ok(cfg) => cfg,
    ///     Err(msg) => {
    ///         eprintln!("{}",msg);
    ///         process::exit(1);
    ///     },
    /// };
    /// ```
    pub fn new(args: Vec<String>) -> Result<Config, String> {
        // setup default values
        let mut stations_file = String::from(DEFAULT_STATIONS_FILE);
        let mut rsd_file = String::from(DEFAULT_RSD_FILE);
        let mut screamer_file = String::from(DEFAULT_SCREAMERSTATIONS_FILE);
        let mut genre_filter_str = String::from("");
        let mut genre_filter_type = FilterType::Nothing;
        let mut nation_filter_str = String::from("");
        let mut nation_filter_type = FilterType::Nothing;

        // analyze and setup values
        let mut args_iter = args.iter();
        args_iter.next(); // 最初は飛ばす
        for arg in args_iter {
            if arg == "-v" {
                Config::show_version();
                return Err(Config::MSG_ERR_VERSION.to_string());
            }
            if arg == "-h" {
                Config::show_help();
                return Err(Config::MSG_ERR_EHELP.to_string());
            };
            if arg == "-hj" {
                Config::show_help_j();
                return Err(Config::MSG_ERR_JHELP.to_string());
            };
            if arg.chars().count() < 5 {
                eprintln!("invalid length prm : {}", arg);
                continue;
            };
            let fourth_ch = arg.char_indices().nth(4).unwrap().0;
            match &arg[0..fourth_ch] {
                "-df=" => {
                    stations_file = String::from(&arg[fourth_ch..]);
                }
                "-rs=" => {
                    rsd_file = String::from(&arg[fourth_ch..]);
                }
                "-ss=" => {
                    screamer_file = String::from(&arg[fourth_ch..]);
                }
                "-gi=" => {
                    genre_filter_type = FilterType::Include;
                    genre_filter_str = String::from(&arg[fourth_ch..]);
                }
                "-ge=" => {
                    if genre_filter_type != FilterType::Include {
                        genre_filter_type = FilterType::Exclude;
                        genre_filter_str = String::from(&arg[fourth_ch..]);
                    };
                }
                "-ni=" => {
                    nation_filter_type = FilterType::Include;
                    nation_filter_str = String::from(&arg[fourth_ch..]);
                }
                "-ne=" => {
                    if nation_filter_type != FilterType::Include {
                        nation_filter_type = FilterType::Exclude;
                        nation_filter_str = String::from(&arg[fourth_ch..]);
                    };
                }
                _ => eprintln!("invalid prm : {}", arg),
            }
        }

        // check and make parameters
        if stations_file == "-" {
            stations_file = String::from("");
        };
        let genre_filter = REFilter::new(genre_filter_type, &genre_filter_str)?;
        let nation_filter = REFilter::new(nation_filter_type, &nation_filter_str)?;

        // show version
        Config::show_version();

        // return Ok(Config)
        Ok(Config {
            stations_file,
            rsd_file,
            screamer_file,
            genre_filter,
            nation_filter,
        })
    }

    pub(crate) const MSG_ERR_VERSION: &str = "show version";
    pub(crate) const MSG_ERR_EHELP: &str = "show help";
    pub(crate) const MSG_ERR_JHELP: &str = "show japanese help";

    /// check normal end or error with msg.
    /// If process is normal end, return true. otherwise return false.
    ///
    /// Config::newのエラーメッセージをチェックして、正常終了か
    /// エラーによる終了かを判別します。
    /// 正常終了ならtrueを、エラーならfalseを返します。
    ///
    /// # Examples
    /// ```
    /// use rs2sr::config::*;
    ///
    /// let args = vec!["dummy".to_string()];
    /// match Config::new(args) {
    ///     Ok(_) => println!("success"),
    ///     Err(msg) =>
    ///         if Config::is_no_error(&msg) {
    ///             println!("success and end");
    ///         } else {
    ///             println!("error");
    ///         },
    /// };
    /// ```
    pub fn is_no_error(msg: &str) -> bool {
        msg == Config::MSG_ERR_VERSION
            || msg == Config::MSG_ERR_EHELP
            || msg == Config::MSG_ERR_JHELP
    }

    /// get Screamer Radio's default stations file name.
    ///
    /// Screamer Radioのデフォルトの局名ファイル名を返します。
    ///
    /// # Examples
    /// ```
    /// use rs2sr::config::*;
    /// use std::process;
    ///
    /// let args: Vec<String> = vec!["dummy".to_string(),
    ///                              "-rs=hogehoge.rsd".to_string(),
    ///                              "-df=nantoka.json".to_string()];
    /// let cfg = match Config::new(args) {
    ///     Ok(cfg) =>  cfg,
    ///     Err(msg) => {
    ///         eprintln!("{}",msg);
    ///         process::exit(1);
    ///     },
    /// };
    /// assert_eq!(cfg.get_stations_filename(),"nantoka.json");
    /// ```
    pub fn get_stations_filename(&self) -> &str {
        &self.stations_file
    }

    /// get Radio Sure!'s stations file name.
    ///
    /// Radio Sure!の局名ファイル名を返します。
    ///
    /// # Examples
    /// ```
    /// use rs2sr::config::*;
    /// use std::process;
    ///
    /// let args: Vec<String> = vec!["dummy".to_string(),
    ///                              "-rs=hogehoge.rsd".to_string(),
    ///                              "-ss=nantoka.json".to_string()];
    /// let cfg = match Config::new(args) {
    ///     Ok(cfg) =>  cfg,
    ///     Err(msg) => {
    ///         eprintln!("{}",msg);
    ///         process::exit(1);
    ///     },
    /// };
    /// assert_eq!(cfg.get_rsd_filename(),"hogehoge.rsd");
    /// ```
    pub fn get_rsd_filename(&self) -> &str {
        &self.rsd_file
    }

    /// get Screamer Radio's user stations file name.
    ///
    /// Screamer Radioのユーザー局名ファイル名を返します。
    ///
    /// # Examples
    /// ```
    /// use rs2sr::config::*;
    /// use std::process;
    ///
    /// let args: Vec<String> = vec!["dummy".to_string(),
    ///                              "-rs=hogehoge.rsd".to_string(),
    ///                              "-ss=nantoka.json".to_string()];
    /// let cfg = match Config::new(args) {
    ///     Ok(cfg) =>  cfg,
    ///     Err(msg) => {
    ///         eprintln!("{}",msg);
    ///         process::exit(1);
    ///     },
    /// };
    /// assert_eq!(cfg.get_screamer_filename(),"nantoka.json");
    /// ```
    pub fn get_screamer_filename(&self) -> &str {
        &self.screamer_file
    }

    /// Determine if the specified genre matches the preset genres.
    /// If genre matches, return true.
    ///
    /// 指定したgenreが事前に設定したジャンルとマッチするかどうか判定します。
    /// マッチするならば true を返します。
    ///
    /// # Examples
    /// ```
    /// use rs2sr::config::*;
    /// use std::process;
    ///
    /// let args: Vec<String> = vec!["dummy".to_string(),
    ///                              "-gi=Jazz,Rock,Blues".to_string()];
    /// let cfg = match Config::new(args) {
    ///     Ok(cfg) => cfg,
    ///     Err(msg) => {
    ///         eprintln!("{}",msg);
    ///         process::exit(1);
    ///     },
    /// };
    /// assert!(cfg.is_match_genre("Jazz"));
    /// assert!(!cfg.is_match_genre("Christmas"));
    /// ```
    pub fn is_match_genre(&self, genre: &str) -> bool {
        self.genre_filter.is_match(genre)
    }

    /// Determine if the specified nation matches the preset nations.
    /// If genre matches, return true.
    ///
    /// 指定した国名が事前に設定した国名とマッチするかどうか判定します。
    /// マッチするならば true を返します。
    ///
    /// # Examples
    /// ```
    /// use rs2sr::config::*;
    /// use std::process;
    ///
    /// let args: Vec<String> = vec!["dummy".to_string(),
    ///                              "-ni=Germany,New Zealand".to_string()];
    /// let cfg = match Config::new(args) {
    ///     Ok(cfg) => cfg,
    ///     Err(msg) => {
    ///         eprintln!("{}",msg);
    ///         process::exit(1);
    ///     },
    /// };
    /// assert!(cfg.is_match_nation("New Zealand"));
    /// assert!(!cfg.is_match_nation("USA"));
    /// ```
    pub fn is_match_nation(&self, nation: &str) -> bool {
        self.nation_filter.is_match(nation)
    }

    // define rs2sr version string fot -v option
    //
    // -vオプション用のバージョン文字列
    const RS2SR_TITLE: &str = "rs2sr version 0.3.0";

    // show rs2sr version
    //
    // rs2srのバージョンを表
    fn show_version() {
        println!("{}", Config::RS2SR_TITLE);
    }

    // show english help
    //
    // 英語版ヘルプ表示
    fn show_help() {
        println!(
            "{}",
            Config::RS2SR_TITLE.to_owned()
                + "
command line options:
-h  ... this help
-hj ... japanese help
-v  ... version
-df=filename ... filename is Screamer Radio's stations data file.
                 default name is \"stations.json\".
                 This file is used to detect duplicate station names.
                 If you use filename \"-\", no check dups.
-rs=filename ... filename is Radio Sure!'s stations data file.
                 default name is \"stations-2022-03-07.rsd\".
-ss=filename ... filename is Screamer Radio's user stations data file.
                 default name is \"ScreamerStations.json\".
-gi=pattern1[,pattern2,...] ... pickup pattern genre stations.
-ge=pattern1[,pattern2,...] ... pickup without pattern genre stations.
                                *IF YOU USE -gi option, IGNORE -ge option*
-ni=pattern1[,pattern2,...] ... pickup pattern nation stations.
-ne=pattern1[,pattern2,...] ... pickup without pattern nation stations.
                                *IF YOU USE -ni option, IGNORE -ne option*

If filename or pattern include space, you use \"\".
ex. -ds=\"something space file.json\"
    -gi=\"Jazz,Adult Contemporary\""
        );
    }

    // show japanese help
    //
    // 日本語版ヘルプ表示
    fn show_help_j() {
        println!(
            "{}",
            Config::RS2SR_TITLE.to_owned()
                + "
コマンドライン オプション：
-h  ... 英語版ヘルプ
-hj ... このヘルプ
-v  ... バージョン表記
-df=filename ... filenameはScreamer Radioのstations data fileです。
                 defaultは \"stations.json\" です。
                 このファイルはRadio Sure!とScreamer Radioの間で重複する名前を
                 チェックするのに使用します。
                 filenameに - を使用した場合、この重複チェックは行われません。
-rs=filename ... filenameはRadio Sure!のstations data fileです。
                 defaultは \"stations-2022-03-07.rsd\" です。
-ss=filename ... filenameはScreamer Radioのuser stations data fileです。
                 defaultは \"ScreamerStations.json\" です。
-gi=pattern1[,pattern2,...] ... 指定したパターンのジャンルを選択します。
-ge=pattern1[,pattern2,...] ... 指定したパターン\"以外\"のジャンルを選択します。
                                なお -gi オプションを使用した場合、 -ge オプシ
                                ョンは無視されます。
-ni=pattern1[,pattern2,...] ... 指定したパターンの国を選択します。
-ne=pattern1[,pattern2,...] ... 指定したパターン\"以外\"の国を選択します。
                                なお -ni オプションを使用した場合、 -ne オプシ
                                ョンは「無視」されます。

もしジャンルや国にスペースが含まれる場合は \"\"でくくってください。
例. -ds=\"something space file.json\"
    -gi=\"Jazz,Adult Contemporary\""
        );
    }
}

//
// 以下テスト
//
#[cfg(test)]
mod tests {
    use super::*;

    #[test]
    fn cfg_refilter_success1() -> Result<(), String> {
        let cfg = REFilter::new(FilterType::Nothing, "foo,bar")?;
        assert!(cfg.is_match("foo"));
        assert!(cfg.is_match("bar"));
        assert!(cfg.is_match("baz"));

        Ok(())
    }

    #[test]
    fn cfg_refilter_success2() -> Result<(), String> {
        let cfg = REFilter::new(FilterType::Include, "foo,bar")?;
        assert!(cfg.is_match("foo"));
        assert!(cfg.is_match("bar"));
        assert!(!cfg.is_match("baz"));

        Ok(())
    }

    #[test]
    fn cfg_refilter_success3() -> Result<(), String> {
        let cfg = REFilter::new(FilterType::Exclude, "foo,bar")?;
        assert!(!cfg.is_match("foo"));
        assert!(!cfg.is_match("bar"));
        assert!(cfg.is_match("baz"));

        Ok(())
    }

    #[test]
    fn cfg_refilter_failure() -> Result<(), String> {
        match REFilter::new(FilterType::Exclude, "") {
            Ok(_) => {
                assert!(false);
                Err("what's happen?".to_string())
            }
            Err(_) => Ok(()),
        }
    }

    #[test]
    fn cfg_config_success1() -> Result<(), String> {
        let args: Vec<String> = vec!["dummy".to_string()];
        let cfg = Config::new(args)?;

        assert_eq!(cfg.get_stations_filename(), DEFAULT_STATIONS_FILE);
        assert_eq!(cfg.get_rsd_filename(), DEFAULT_RSD_FILE);
        assert_eq!(cfg.get_screamer_filename(), DEFAULT_SCREAMERSTATIONS_FILE);
        assert!(cfg.is_match_genre("hoge"));
        assert!(cfg.is_match_genre("fuga"));
        assert!(cfg.is_match_nation("hoge"));
        assert!(cfg.is_match_nation("fuga"));

        Ok(())
    }

    #[test]
    fn cfg_config_success2() -> Result<(), String> {
        let args: Vec<String> = vec![
            "dummy".to_string(),
            "-df=tekitou.json".to_string(),
            "-rs=hogehoge.rsd".to_string(),
            "-ss=nantoka.json".to_string(),
            "-gi=Jazz,Rock".to_string(),
            "-ni=Germany,UK".to_string(),
        ];
        let cfg = Config::new(args)?;

        assert_eq!(cfg.get_stations_filename(), "tekitou.json");
        assert_eq!(cfg.get_rsd_filename(), "hogehoge.rsd");
        assert_eq!(cfg.get_screamer_filename(), "nantoka.json");
        assert!(cfg.is_match_genre("Jazz"));
        assert!(cfg.is_match_genre("Rock"));
        assert!(!cfg.is_match_genre("Blues"));
        assert!(cfg.is_match_nation("Germany"));
        assert!(cfg.is_match_nation("UK"));
        assert!(!cfg.is_match_nation("USA"));

        Ok(())
    }

    #[test]
    fn cfg_config_success3() -> Result<(), String> {
        let args: Vec<String> = vec![
            "dummy".to_string(),
            "-df=-".to_string(),
            "-ge=Jazz,Rock".to_string(),
            "-ne=Germany,UK".to_string(),
        ];
        let cfg = Config::new(args)?;

        assert_eq!(cfg.get_stations_filename(), "");
        assert!(!cfg.is_match_genre("Jazz"));
        assert!(!cfg.is_match_genre("Rock"));
        assert!(cfg.is_match_genre("Blues"));
        assert!(!cfg.is_match_nation("Germany"));
        assert!(!cfg.is_match_nation("UK"));
        assert!(cfg.is_match_nation("USA"));

        Ok(())
    }

    #[test]
    fn cfg_config_success4() -> Result<(), String> {
        let args: Vec<String> = vec![
            "dummy".to_string(),
            "-df=".to_string(),
            "-rs=".to_string(),
            "-ss=".to_string(),
            "-gi=Jazz,Rock".to_string(),
            "-ni=Germany,UK".to_string(),
            "-ge=Rock,Classic".to_string(),
            "-ne=UK,USA".to_string(),
        ];
        let cfg = Config::new(args)?;

        assert!(cfg.is_match_genre("Jazz"));
        assert!(cfg.is_match_genre("Rock"));
        assert!(!cfg.is_match_genre("Blues"));
        assert!(cfg.is_match_nation("Germany"));
        assert!(cfg.is_match_nation("UK"));
        assert!(!cfg.is_match_nation("Japan"));

        Ok(())
    }

    #[test]
    fn cfg_config_success5() -> Result<(), String> {
        let args: Vec<String> = vec![
            "dummy".to_string(),
            "hogehoge".to_string(),
            "aa".to_string(),
        ];
        let cfg = Config::new(args)?;

        assert_eq!(cfg.get_stations_filename(), DEFAULT_STATIONS_FILE);
        assert_eq!(cfg.get_rsd_filename(), DEFAULT_RSD_FILE);
        assert_eq!(cfg.get_screamer_filename(), DEFAULT_SCREAMERSTATIONS_FILE);
        assert!(cfg.is_match_genre("hoge"));
        assert!(cfg.is_match_genre("fuga"));
        assert!(cfg.is_match_nation("hoge"));
        assert!(cfg.is_match_nation("fuga"));

        Ok(())
    }

    #[test]
    fn cfg_config_failure1() -> Result<(), String> {
        let args: Vec<String> = vec![
            "dummy".to_string(),
            "-gi=Jazz".to_string(),
            "-v".to_string(),
        ];
        match Config::new(args) {
            Ok(_) => return Err("what's happen?".to_string()),
            Err(msg) => assert_eq!(msg, Config::MSG_ERR_VERSION),
        };
        Ok(())
    }

    #[test]
    fn cfg_config_failure2() -> Result<(), String> {
        let args: Vec<String> = vec![
            "dummy".to_string(),
            "-gi=Jazz".to_string(),
            "-h".to_string(),
        ];
        match Config::new(args) {
            Ok(_) => return Err("what's happen?".to_string()),
            Err(msg) => assert_eq!(msg, Config::MSG_ERR_EHELP),
        };
        Ok(())
    }

    #[test]
    fn cfg_config_failure3() -> Result<(), String> {
        let args: Vec<String> = vec![
            "dummy".to_string(),
            "-gi=Jazz".to_string(),
            "-hj".to_string(),
        ];
        match Config::new(args) {
            Ok(_) => return Err("what's happen?".to_string()),
            Err(msg) => assert_eq!(msg, Config::MSG_ERR_JHELP),
        };
        Ok(())
    }

    #[test]
    fn cfg_is_no_error_success1() {
        assert!(Config::is_no_error(Config::MSG_ERR_VERSION));
    }

    #[test]
    fn cfg_is_no_error_success2() {
        assert!(Config::is_no_error(Config::MSG_ERR_EHELP));
    }

    #[test]
    fn cfg_is_no_error_success3() {
        assert!(Config::is_no_error(Config::MSG_ERR_JHELP));
    }

    #[test]
    fn cfg_is_no_error_failure() {
        assert!(!Config::is_no_error("hogehoge!"));
    }
}
