//! # converter Crate
//!
//! `converter' Crate provides convert function from Radio Sure! data to
//! Screamer Radio data.
//!
//! このCrateはRadio Sure!のデータをScreamer Radioのデータに変換する関数
//! を提供します。

use regex::Regex;
use uuid::Uuid;
use validator;
use crate::radiosure::*;
use crate::screamerradio::userdatas::*;

// lineからurlを見つけ出して返す。複数のurlが存在する場合は最初の
// ものを採用する。
// 見つけたurlが不正だった場合、又は、urlが見つからなかった場合は
// http://127.0.0.1/を返す。
// なお re は事前にurlにマッチする正規表現を指定してある。
const LOCAL_URL: &str = "http://127.0.0.1/";

fn get_url(line: &str, re: &regex::Regex) -> String {
    // 戻り値を準備。とりあえずはdummy値を設定しておく
    let mut result = String::from(LOCAL_URL);

    // マッチングした順番にチェックして正しいURLならそれを返す
    for caps in re.captures_iter(line) {
        let url = String::from(&caps[0]);
        if validator::validate_url(&url) {
            result = url;
            break;
        };
    }

    result
}

// RsdataをSrdataに変換する。
// なお re は事前にurlにマッチする正規表現を指定してある。
fn rsdata_to_srdata(rsdata: Rsdata, re: &regex::Regex) -> Srdata {
    // rsdataから必要なデータを作成しておく
    let id = String::from("user:") + &Uuid::new_v4().to_string();
    let title = rsdata.station_name;
    let website_url = get_url(&rsdata.note, re);
    let mut tags: Vec<String> = Vec::new();
    if !rsdata.tag.is_empty() {
        tags.push(rsdata.tag);
    }
    if !rsdata.nation.is_empty() {
        tags.push(rsdata.nation);
    }
    tags.push("User".to_string());

    Srdata {
        id,
        title,
        website_url,
        tags,
        sources: rsdata.sources,
    }
}

const RE_URL_STR: &str = r"(https?://[\w/:;%#\$&\?\(\)~\.=\+\-_]+)";

/// Convert `Vec<Rsdata>` to `Vec<Srdata>`
///
/// `Vec<Rsdata>`を`Vec<Srdata>`に変換します。
///
/// # Examples
///
/// ```
/// use rs2sr::radiosure::*;
/// use rs2sr::screamerradio::userdatas::*;
/// use rs2sr::converter::*;
///
/// let rsdatas:Vec<Rsdata> = Vec::new();
/// let srdatas = rsdatas_to_srdatas(rsdatas);
/// ```
pub fn rsdatas_to_srdatas(rsdatas: Vec<Rsdata>) -> Vec<Srdata> {
    println!("convert struct to struct...");
    let mut srdatas: Vec<Srdata> = Vec::new();

    // 毎回生成するとえらい時間がかかるので、loop前に作成して
    // 参照を渡すようにする
    let re_url_str = RE_URL_STR;
    let re = Regex::new(re_url_str).unwrap();

    for rsdata in rsdatas {
        srdatas.push(rsdata_to_srdata(rsdata, &re));
    }

    srdatas
}

//
// 以下テスト
//
#[cfg(test)]
mod tests {
    use super::*;

    #[test]
    fn sr_get_url_success() {
        let re_url_str = RE_URL_STR;
        let re = Regex::new(re_url_str).unwrap();

        let http_str = "http://www.example.com:8080/";
        let https_str = "https://special.example.com/hogehoge";
        let https_str2 = "https://special.example.com:80/;";

        let mod_http_str = String::from("these strs are noise ")
            + http_str
            + &String::from(" ;; these strs are noise.");
        let mod_https_str = String::from("these strs are noise ")
            + https_str
            + &String::from(" ;; these strs are noise.");
        let mod_https_str2 = String::from("these strs are noise ")
            + https_str2
            + &String::from(" ;; these strs are noise.");

        assert_eq!(get_url(&mod_http_str, &re), http_str);
        assert_eq!(get_url(&mod_https_str, &re), https_str);
        assert_eq!(get_url(&mod_https_str2, &re), https_str2);
    }

    #[test]
    fn sr_get_url_failure() {
        let re_url_str = RE_URL_STR;
        let re = Regex::new(re_url_str).unwrap();

        let local_http = LOCAL_URL;

        let http_str = "http://www.example.com:gdgd/";
        let https_str = "https://%20special/";

        let mod_http_str = String::from("these strs are noise ")
            + http_str
            + &String::from(" ;; these strs are noise.");
        let mod_https_str = String::from("these strs are noise ")
            + https_str
            + &String::from(" ;; these strs are noise.");

        assert_eq!(get_url(&mod_http_str, &re), local_http);
        assert_eq!(get_url(&mod_https_str, &re), local_http);
    }

    #[test]
    fn sr_rsdata_to_srdata_success1() {
        let re_url_str = RE_URL_STR;
        let re = Regex::new(re_url_str).unwrap();
        // tag,nation有り、noteにurl無し
        let rsdata = Rsdata {
            station_name: String::from("hogehoge"),
            note: String::from("tekito- na mojiretu"),
            tag: String::from("Jazz"),
            nation: String::from("USA"),
            _language: String::from("English"),
            sources: vec![
                "https://www.example.com/jazz".to_string(),
                "https://www.example.com:8080/jazz".to_string(),
            ],
        };
        let right_data = Srdata {
            id: "user:uuid_v4_tekitou01".to_string(),
            title: "hogehoge".to_string(),
            website_url: LOCAL_URL.to_string(),
            tags: vec!["Jazz".to_string(), "USA".to_string(), "User".to_string()],
            sources: vec![
                "https://www.example.com/jazz".to_string(),
                "https://www.example.com:8080/jazz".to_string(),
            ],
        };

        let srdata = rsdata_to_srdata(rsdata, &re);

        assert!(srdata.loosy_eq(&right_data));
    }

    #[test]
    fn sr_rsdata_to_srdata_success2() {
        let re_url_str = RE_URL_STR;
        let re = Regex::new(re_url_str).unwrap();
        // tag,nation無し、noteにurlが一つ
        let rsdata = Rsdata {
            station_name: String::from("hogehoge"),
            note: String::from("tekito- na mojiretu https://stream.example.com/ nado"),
            tag: String::from(""),
            nation: String::from(""),
            _language: String::from("English"),
            sources: vec![
                "https://www.example.com/jazz".to_string(),
                "https://www.example.com:8080/jazz".to_string(),
            ],
        };
        let right_data = Srdata {
            id: "user:uuid_v4_tekitou01".to_string(),
            title: "hogehoge".to_string(),
            website_url: "https://stream.example.com/".to_string(),
            tags: vec!["User".to_string()],
            sources: vec![
                "https://www.example.com/jazz".to_string(),
                "https://www.example.com:8080/jazz".to_string(),
            ],
        };

        let srdata = rsdata_to_srdata(rsdata, &re);

        assert!(srdata.loosy_eq(&right_data));
    }

    #[test]
    fn sr_rsdata_to_srdata_success3() {
        let re_url_str = RE_URL_STR;
        let re = Regex::new(re_url_str).unwrap();
        // tag無し,nation有り、noteにurlが二つ
        let rsdata = Rsdata {
            station_name: String::from("hogehoge"),
            note: String::from("tekito- na mojiretu https://stream.example.com/1 https://stream.example.com/2 nado"),
            tag: String::from(""),
            nation: String::from("USA"),
            _language: String::from("English"),
            sources: vec!["https://www.example.com/jazz".to_string(),
                          "https://www.example.com:8080/jazz".to_string()],
        };
        let right_data = Srdata {
            id: "user:uuid_v4_tekitou01".to_string(),
            title: "hogehoge".to_string(),
            website_url: "https://stream.example.com/1".to_string(),
            tags: vec!["USA".to_string(), "User".to_string()],
            sources: vec![
                "https://www.example.com/jazz".to_string(),
                "https://www.example.com:8080/jazz".to_string(),
            ],
        };

        let srdata = rsdata_to_srdata(rsdata, &re);

        assert!(srdata.loosy_eq(&right_data));
    }

    #[test]
    fn sr_rsdata_to_srdata_success4() {
        let re_url_str = RE_URL_STR;
        let re = Regex::new(re_url_str).unwrap();
        // tag有り,nation無し、noteに無効なurlが一つ
        let rsdata = Rsdata {
            station_name: String::from("hogehoge"),
            note: String::from("tekito- na mojiretu https://stream.example.com:boge/1 https://stream.example.com/2 nado"),
            tag: String::from("Jazz"),
            nation: String::from(""),
            _language: String::from("English"),
            sources: vec!["https://www.example.com/jazz".to_string(),
                          "https://www.example.com:8080/jazz".to_string()],
        };
        let right_data = Srdata {
            id: "user:uuid_v4_tekitou01".to_string(),
            title: "hogehoge".to_string(),
            website_url: "https://stream.example.com/2".to_string(),
            tags: vec!["Jazz".to_string(), "User".to_string()],
            sources: vec![
                "https://www.example.com/jazz".to_string(),
                "https://www.example.com:8080/jazz".to_string(),
            ],
        };

        let srdata = rsdata_to_srdata(rsdata, &re);

        assert!(srdata.loosy_eq(&right_data));
    }

    #[test]
    fn sr_rsdatas_to_srdatas_success() {
        let rsdatas = vec![
            Rsdata {
                station_name: String::from("hogehoge"),
                note: String::from("tekito- na mojiretu https://stream.example.com/ nado"),
                tag: String::from(""),
                nation: String::from(""),
                _language: String::from("English"),
                sources: vec![
                    "https://www.example.com/jazz".to_string(),
                    "https://www.example.com:8080/jazz".to_string(),
                ],
            },
            Rsdata {
                station_name: String::from("fugafuga"),
                note: String::from("tekito- na mojiretu https://boke.example.com/; nado"),
                tag: String::from("Pop"),
                nation: String::from("New Zealand"),
                _language: String::from("English"),
                sources: vec![
                    "https://boke.example.com/;".to_string(),
                    "https://boke.example.com:8080/pop".to_string(),
                ],
            },
            Rsdata {
                station_name: String::from("mocho mocho"),
                note: String::from("tekito- na mojiretu https://mochi.example.com:boge/; https://mochimochi.example.com:8080/?stat=neko nado"),
                tag: String::from("Jazz"),
                nation: String::from(""),
                _language: String::from(""),
                sources: vec![
                    "https://mochi.example.com:8080/".to_string(),
                ],
            },
        ];
        let right_datas = vec![
            Srdata {
                id: "user:uuid_v4_tekitou01".to_string(),
                title: "hogehoge".to_string(),
                website_url: "https://stream.example.com/".to_string(),
                tags: vec!["User".to_string()],
                sources: vec![
                    "https://www.example.com/jazz".to_string(),
                    "https://www.example.com:8080/jazz".to_string(),
                ],
            },
            Srdata {
                id: "user:uuid_v4_tekitou02".to_string(),
                title: "fugafuga".to_string(),
                website_url: "https://boke.example.com/;".to_string(),
                tags: vec![
                    "Pop".to_string(),
                    "New Zealand".to_string(),
                    "User".to_string(),
                ],
                sources: vec![
                    "https://boke.example.com/;".to_string(),
                    "https://boke.example.com:8080/pop".to_string(),
                ],
            },
            Srdata {
                id: "user:uuid_v4_tekitou03".to_string(),
                title: "mocho mocho".to_string(),
                website_url: "https://mochimochi.example.com:8080/?stat=neko".to_string(),
                tags: vec!["Jazz".to_string(), "User".to_string()],
                sources: vec!["https://mochi.example.com:8080/".to_string()],
            },
        ];

        let sr_datas = rsdatas_to_srdatas(rsdatas);

        assert!(sr_datas.len() == right_datas.len(), "differnt data lens.");
        let mut r_iter = right_datas.iter();
        for l in sr_datas {
            let r = r_iter.next().unwrap();
            assert!(l.loosy_eq(&r));
        }
    }
}
