//! # radiosure Crate
//!
//! `radiosure` is library for reading Radio Sure! data and making structs of
//! these datas. When making structs, filter with tags and nation names.
//! 
//! このCrateはRadio Sure!のデフォルトのstation data fileを読み込んで構造体
//! に格納します。その際、タグや国名でフィルタリングして幾つか除外してたりも
//! します。

use crate::config::*;
use crate::screamerradio::defaultdatas::*;
use csv::StringRecord;
use std::fs::File;
use validator;

/// Radio Sure! station data structure. All params are limited pub for crate.
///
/// Radio Sure!の局データを保持する構造体です。メンバーは全部crate限定公開と
/// なります。
#[derive(Debug, PartialEq, Eq)]
pub struct Rsdata {
    pub(crate) station_name: String, // 局名
    pub(crate) note: String,         // 記事。存在するならばここからurlを抜き出す
    pub(crate) tag: String,          // タグ(複数)
    pub(crate) nation: String,       // 国名
    pub(crate) _language: String,    // 言語(未使用)
    pub(crate) sources: Vec<String>, // streaming用url。最大６個まで
}

// station.sourcesが"-"だったら空文字列を返却する。
fn check_and_replace_minus(word: &str) -> String {
    if word == "-" {
        String::from("")
    } else {
        String::from(word)
    }
}

// TSVの行からRadio Sure!のデータに変換する。
// 成功すればSome(Rsdata)、失敗すればNoneを返す。
fn convert_strs_to_rsdata(datas: &StringRecord, cfg: &Config, line_no: usize) -> Option<Rsdata> {
    // precheck data field nums
    if datas.len() != 11 {
        eprintln!(
            "(conv tsv to rsdata) prm count : {} / line.{}",
            datas.len(),
            line_no
        );
        return None;
    }

    // radio station name
    let station_name = check_and_replace_minus(datas.get(0).unwrap());
    if station_name.is_empty() {
        eprintln!(
            "(conv tsv to rsdata) station name is none / line.{}",
            line_no
        );
        return None;
    }

    // note
    let note = check_and_replace_minus(datas.get(1).unwrap());

    // Tag
    let tag = check_and_replace_minus(datas.get(2).unwrap());
    if !cfg.is_match_genre(&tag) {
        // マッチしないジャンルは除外
        return None;
    }

    // 国名
    let nation = check_and_replace_minus(datas.get(3).unwrap());
    if !cfg.is_match_nation(&nation) {
        // マッチしない国名は除外
        return None;
    }

    // 言語
    let language = check_and_replace_minus(datas.get(4).unwrap());

    // stream resources
    let mut sources = Vec::<String>::new();
    for index in 5..11 {
        let source = check_and_replace_minus(datas.get(index).unwrap());
        if !source.is_empty() {
            if validator::validate_url(&source) {
                sources.push(source);
            } else {
                eprintln!("{}", station_name);
                eprintln!("  invalid url => {}", source);
            };
        }
    }
    if sources.is_empty() {
        // stream sourceが一つも無い！
        eprintln!("(conv tsv to rsdata) no stream resource / line.{}", line_no);
        return None;
    }

    // Radio Sureのデータを作成
    let result = Rsdata {
        station_name,
        note,
        tag,
        nation,
        _language: language,
        sources,
    };
    Some(result)
}

/// # convert_file_to_rsdatas
///
/// Convert Radio Sure! station data to struct from Radio Sure! station file
/// without stations what is already exist on DefaultStations checked by 
/// DefaultStationChecker.
/// If successed, return Ok(`Vec<Rsdata>`).
/// If failure, return Err(String).
///
/// Radio Sure!の局名ファイルを読み込んでデータを作成します。その際に
/// DefaultStationCheckerを参照して、既に存在する局名だった場合は除外します。
/// 関数が成功した場合はOk(`Vec<Rsdata>`)を返します。
/// 失敗した場合はErr(String)を返します。
///
/// # Examples
///
/// ```
/// use std::process::exit;
/// use rs2sr::config::*;
/// use rs2sr::screamerradio::defaultdatas::*;
/// use rs2sr::radiosure::*;
///
/// let args:Vec<String> = vec!["dummy".to_string(),
///                             "-rs=tests/datas/sonzai_sinai.rsd".to_string(),
///                            ];
/// let cfg = match Config::new(args) {
///     Ok(cfg) => cfg,
///     Err(msg) => {
///       eprintln!("{}",msg);
///       return ;
///     },
/// };
/// let dsc = match DefaultStationChecker::new("") {
///     Ok(dsc) => dsc,
///     Err(msg) => {
///         eprintln!("{}",msg);
///         return ;
///     },
/// };
///
/// let rsdatas = match convert_file_to_rsdatas(&cfg, dsc) {
///    Ok(rsdatas) => rsdatas,
///    Err(msg) => {
///      eprintln!("something bad : {}", msg);
///      return ;
///    },
/// };
/// ```
pub fn convert_file_to_rsdatas(
    cfg: &Config,
    dsc: DefaultStationChecker,
) -> Result<Vec<Rsdata>, String> {
    // ファイルを開く
    println!("read Radio Sure! data...");
    let f = match File::open(cfg.get_rsd_filename()) {
        Err(e) => return Err(String::from("file open/") + &e.to_string()),
        Ok(file) => file,
    };

    // tsv readerbuilderを準備
    let mut tsv_reader = csv::ReaderBuilder::new()
        .has_headers(false) // headerは無い！
        .delimiter(b'\t') // 区切りはtabに変更
        .from_reader(f); // ファイルから読み込む

    // Radio Sure!のデータ準備
    let mut rsdatas: Vec<Rsdata> = Vec::new();

    // 重複したステーション名の保存
    let mut dup_stations: Vec<String> = Vec::new();

    // Raido Sure!のデータに変換
    println!("convert tsv to struct...");
    let mut line_count = 0;
    for result in tsv_reader.records() {
        if let Ok(datas) = result {
            if let Some(rsdata) = convert_strs_to_rsdata(&datas, cfg, line_count + 1) {
                if dsc.is_unique(&rsdata.station_name) {
                    // defaultに存在しなかったので追加
                    rsdatas.push(rsdata);
                } else {
                    // defaultに既に存在する
                    dup_stations.push(rsdata.station_name);
                };
            };
        } else {
            eprintln!("illigal tsv data? : line no.{}", line_count + 1);
        };
        line_count += 1;
    }
    println!("all stations's data ...... {}", line_count);
    println!("    found new stations ... {}", rsdatas.len());
    println!("    already exists ....... {}", dup_stations.len());

    // データを返す
    Ok(rsdatas)
}

//
// 以下テスト
//
#[cfg(test)]
mod tests {
    use super::*;

    #[test]
    fn rs_check_and_replace_minus_case1() {
        let teststr = "hogehoge";

        let result = check_and_replace_minus(teststr);
        assert_eq!(result, teststr);
    }

    #[test]
    fn rs_check_and_replace_minus_case2() {
        let teststr = "-";

        let result = check_and_replace_minus(teststr);
        assert!(result.is_empty());
    }

    #[test]
    fn rs_convert_strs_to_rsdata_success1() {
        let args: Vec<String> = vec!["dummy".to_string()];
        let cfg = match Config::new(args) {
            Ok(cfg) => cfg,
            Err(_) => {
                assert!(false);
                return;
            }
        };

        let base_data = vec![
            "radio station",
            "station's note",
            "tag",
            "nation",
            "language",
            "http://example.com/source1",
            "http://example.com/source2",
            "http://example.com/source3",
            "https://example.com/source4",
            "https://example.com/source5",
            "https://example.com/source6",
        ];

        let data = csv::StringRecord::from(base_data.clone());

        let rsdata = convert_strs_to_rsdata(&data, &cfg, 1);
        match rsdata {
            Some(rsdata) => {
                assert_eq!(rsdata.station_name, base_data[0]);
                assert_eq!(rsdata.note, base_data[1]);
                assert_eq!(rsdata.tag, base_data[2]);
                assert_eq!(rsdata.nation, base_data[3]);
                assert_eq!(rsdata._language, base_data[4]);
                assert_eq!(rsdata.sources.len(), 6);
                assert_eq!(rsdata.sources[0], base_data[5]);
                assert_eq!(rsdata.sources[1], base_data[6]);
                assert_eq!(rsdata.sources[2], base_data[7]);
                assert_eq!(rsdata.sources[3], base_data[8]);
                assert_eq!(rsdata.sources[4], base_data[9]);
                assert_eq!(rsdata.sources[5], base_data[10]);
            }
            None => assert!(false), // 成功しているはず
        }
    }

    #[test]
    fn rs_convert_strs_to_rsdata_success2() {
        let args: Vec<String> = vec!["dummy".to_string()];
        let cfg = match Config::new(args) {
            Ok(cfg) => cfg,
            Err(_) => {
                assert!(false);
                return;
            }
        };

        let base_data = vec![
            "radio station",
            "station's note",
            "-",
            "-",
            "-",
            "http://example.com/source1",
            "http://example.com/source2",
            "--",
            "https://example.com/source3",
            "-",
            "-",
        ];

        let data = csv::StringRecord::from(base_data.clone());

        let rsdata = convert_strs_to_rsdata(&data, &cfg, 1);
        match rsdata {
            Some(rsdata) => {
                assert_eq!(rsdata.station_name, base_data[0]);
                assert_eq!(rsdata.note, base_data[1]);
                assert_eq!(rsdata.tag, "");
                assert_eq!(rsdata.nation, "");
                assert_eq!(rsdata._language, "");
                assert_eq!(rsdata.sources.len(), 3);
                assert_eq!(rsdata.sources[0], base_data[5]);
                assert_eq!(rsdata.sources[1], base_data[6]);
                assert_eq!(rsdata.sources[2], base_data[8]);
            }
            None => assert!(false), // 成功しているはず
        }
    }

    // リソースが一つ足りない
    #[test]
    fn rs_convert_strs_to_rsdata_failure1() {
        let args: Vec<String> = vec!["dummy".to_string()];
        let cfg = match Config::new(args) {
            Ok(cfg) => cfg,
            Err(_) => {
                assert!(false);
                return;
            }
        };

        let base_data = vec![
            "radio station",
            "station's note",
            "tag",
            "nation",
            "language",
            "http://example.com/source1",
            "http://example.com/source2",
            "http://example.com/source3",
            "https://example.com/source4",
            "https://example.com/source5",
        ];

        let data = csv::StringRecord::from(base_data.clone());

        let rsdata = convert_strs_to_rsdata(&data, &cfg, 1);
        match rsdata {
            Some(_rsdata) => {
                assert!(false); // 失敗しているはず
            }
            None => assert!(true), // Noneのハズ
        }
    }

    // ステーション名が空
    #[test]
    fn rs_convert_strs_to_rsdata_failure2() {
        let args: Vec<String> = vec!["dummy".to_string()];
        let cfg = match Config::new(args) {
            Ok(cfg) => cfg,
            Err(_) => {
                assert!(false);
                return;
            }
        };

        let base_data = vec![
            "",
            "station's note",
            "tag",
            "nation",
            "language",
            "http://example.com/source1",
            "http://example.com/source2",
            "http://example.com/source3",
            "https://example.com/source4",
            "https://example.com/source5",
            "https://example.com/source6",
        ];

        let data = csv::StringRecord::from(base_data.clone());

        let rsdata = convert_strs_to_rsdata(&data, &cfg, 1);
        match rsdata {
            Some(_rsdata) => {
                assert!(false); // 失敗しているはず
            }
            None => assert!(true), // Noneのハズ
        }
    }

    // ジャンルチェックでデータがはじかれる・その１
    #[test]
    fn rs_convert_strs_to_rsdata_failure3() {
        let args: Vec<String> = vec![
            "dummy".to_string(),
            "-ge=Talk,News,Public".to_string(),
            "-ne=Russian".to_string(),
        ];
        let cfg = match Config::new(args) {
            Ok(cfg) => cfg,
            Err(_) => {
                assert!(false);
                return;
            }
        };

        let base_data = vec![
            "station name",
            "station's note",
            "Talk",
            "nation",
            "language",
            "http://example.com/source1",
            "http://example.com/source2",
            "http://example.com/source3",
            "https://example.com/source4",
            "https://example.com/source5",
            "https://example.com/source6",
        ];

        let data = csv::StringRecord::from(base_data.clone());

        let rsdata = convert_strs_to_rsdata(&data, &cfg, 1);
        match rsdata {
            Some(_rsdata) => {
                assert!(false); // 失敗しているはず
            }
            None => assert!(true), // Noneのハズ
        }
    }

    // ジャンルチェックではじかれる・その２
    #[test]
    fn rs_convert_strs_to_rsdata_failure4() {
        let args: Vec<String> = vec![
            "dummy".to_string(),
            "-ge=Talk,News,Public".to_string(),
            "-ne=Russian".to_string(),
        ];
        let cfg = match Config::new(args) {
            Ok(cfg) => cfg,
            Err(_) => {
                assert!(false);
                return;
            }
        };

        let base_data = vec![
            "station name",
            "station's note",
            "Public",
            "nation",
            "language",
            "http://example.com/source1",
            "http://example.com/source2",
            "http://example.com/source3",
            "https://example.com/source4",
            "https://example.com/source5",
            "https://example.com/source6",
        ];

        let data = csv::StringRecord::from(base_data.clone());

        let rsdata = convert_strs_to_rsdata(&data, &cfg, 1);
        match rsdata {
            Some(_rsdata) => {
                assert!(false); // 失敗しているはず
            }
            None => assert!(true), // Noneのハズ
        }
    }

    // ジャンルチェックではじかれる・その３
    #[test]
    fn rs_convert_strs_to_rsdata_failure5() {
        let args: Vec<String> = vec![
            "dummy".to_string(),
            "-ge=Talk,News,Public".to_string(),
            "-ne=Russian".to_string(),
        ];
        let cfg = match Config::new(args) {
            Ok(cfg) => cfg,
            Err(_) => {
                assert!(false);
                return;
            }
        };

        let base_data = vec![
            "station name",
            "station's note",
            "News",
            "nation",
            "language",
            "http://example.com/source1",
            "http://example.com/source2",
            "http://example.com/source3",
            "https://example.com/source4",
            "https://example.com/source5",
            "https://example.com/source6",
        ];

        let data = csv::StringRecord::from(base_data.clone());

        let rsdata = convert_strs_to_rsdata(&data, &cfg, 1);
        match rsdata {
            Some(_rsdata) => {
                assert!(false); // 失敗しているはず
            }
            None => assert!(true), // Noneのハズ
        }
    }

    // 国名ではじかれる
    #[test]
    fn rs_convert_strs_to_rsdata_failure6() {
        let args: Vec<String> = vec![
            "dummy".to_string(),
            "-ge=Talk,News,Public".to_string(),
            "-ne=Russian".to_string(),
        ];
        let cfg = match Config::new(args) {
            Ok(cfg) => cfg,
            Err(_) => {
                assert!(false);
                return;
            }
        };

        let base_data = vec![
            "station name",
            "station's note",
            "tags",
            "Russian",
            "language",
            "http://example.com/source1",
            "http://example.com/source2",
            "http://example.com/source3",
            "https://example.com/source4",
            "https://example.com/source5",
            "https://example.com/source6",
        ];

        let data = csv::StringRecord::from(base_data.clone());

        let rsdata = convert_strs_to_rsdata(&data, &cfg, 1);
        match rsdata {
            Some(_rsdata) => {
                assert!(false); // 失敗しているはず
            }
            None => assert!(true), // Noneのハズ
        }
    }

    // ストリームが０ではじかれる
    #[test]
    fn rs_convert_strs_to_rsdata_failure7() {
        let args: Vec<String> = vec!["dummy".to_string()];
        let cfg = match Config::new(args) {
            Ok(cfg) => cfg,
            Err(_) => {
                assert!(false);
                return;
            }
        };

        let base_data = vec![
            "station name",
            "station's note",
            "News",
            "nation",
            "language",
            "",
            "",
            "",
            "",
            "",
            "",
        ];

        let data = csv::StringRecord::from(base_data.clone());

        let rsdata = convert_strs_to_rsdata(&data, &cfg, 1);
        match rsdata {
            Some(_rsdata) => {
                assert!(false); // 失敗しているはず
            }
            None => assert!(true), // Noneのハズ
        }
    }

    // フィルター指定無し
    #[test]
    fn rs_convert_file_to_rsdatas_success1() {
        // config準備
        let args: Vec<String> = vec![
            "dummy".to_string(),
            "-rs=tests/datas/radiosure_success1.rsd".to_string(),
        ];
        let cfg = match Config::new(args) {
            Ok(cfg) => cfg,
            Err(_) => {
                assert!(false);
                return;
            }
        };
        // デフォルトデータ(空)準備
        let dsc = match DefaultStationChecker::new("") {
            Ok(dsc) => dsc,
            Err(_) => {
                assert!(false);
                return;
            }
        };

        // 比較用データ準備
        let org_datas = vec![
            Rsdata {
                station_name :  String::from("rs sample 1"),
                note : String::from("this is dummy data."),
                tag : String::from("Community"),
                nation : String::from("Ireland"),
                _language : String::from("English"),
                sources : vec![String::from("https://tekitou.example.com/stream1"),
                               String::from("https://tekitou.example.com/stream2"),
                               String::from("https://tekitou.example.com/stream3"),
                               String::from("https://tekitou.example.com/stream4"),],
            },
            Rsdata {
                station_name :  String::from("rs sample 2"),
                note : String::from("this is dummy data with url. https://hogehoge.example.com:8080/"),
                tag : String::from("Hard Rock"),
                nation : String::from("New Zealand"),
                _language : String::from("English"),
                sources : vec![String::from("https://hogehoge.example.com/stream1"),
                               String::from("https://hogehoge.example.com/stream2"),
                               String::from("https://hogehoge.example.com/stream3"),
                               String::from("https://hogehoge.example.com/stream4"),
                               String::from("https://hogehoge.example.com/stream5"),
                               String::from("https://hogehoge.example.com/stream6"),],
            },
            Rsdata {
                station_name :  String::from("rs sample 3"),
                note : String::from("this is dummy data with two urls. https://nantoka.example.com:8080/ https://www.examples.com/nantoka"),
                tag : String::from("Talk-Sports"),
                nation : String::from("Japan"),
                _language : String::from("Japanese"),
                sources : vec![String::from("https://nantoka.exsample.com/j"),],
            },
            Rsdata {
                station_name :  String::from("dummy radio 4"),
                note : String::from("this is dummy data with url. http://arienai.example.com/"),
                tag : String::from("Hard Rock"),
                nation : String::from("Japan"),
                _language : String::from("Japanese"),
                sources : vec![String::from("http://arienai.example.com/stream"),],
            },
        ];

        match convert_file_to_rsdatas(&cfg, dsc) {
            Ok(rsdatas) => assert_eq!(rsdatas, org_datas),
            Err(_) => assert!(false),
        }
    }

    // フィルター無し＆デフォルト有り
    #[test]
    fn rs_convert_file_to_rsdatas_success2() {
        // config準備
        let args: Vec<String> = vec![
            "dummy".to_string(),
            "-rs=tests/datas/radiosure_success1.rsd".to_string(),
            "-df=tests/datas/DefaultStations_success1.json".to_string(),
        ];
        let cfg = match Config::new(args) {
            Ok(cfg) => cfg,
            Err(_) => {
                assert!(false);
                return;
            }
        };

        // デフォルトデータ準備
        let dsc = match DefaultStationChecker::new(cfg.get_stations_filename()) {
            Ok(dsc) => dsc,
            Err(_) => {
                assert!(false);
                return;
            }
        };

        // 比較用データ準備
        let org_datas = vec![
            Rsdata {
                station_name :  String::from("rs sample 1"),
                note : String::from("this is dummy data."),
                tag : String::from("Community"),
                nation : String::from("Ireland"),
                _language : String::from("English"),
                sources : vec![String::from("https://tekitou.example.com/stream1"),
                               String::from("https://tekitou.example.com/stream2"),
                               String::from("https://tekitou.example.com/stream3"),
                               String::from("https://tekitou.example.com/stream4"),],
            },
            Rsdata {
                station_name :  String::from("rs sample 2"),
                note : String::from("this is dummy data with url. https://hogehoge.example.com:8080/"),
                tag : String::from("Hard Rock"),
                nation : String::from("New Zealand"),
                _language : String::from("English"),
                sources : vec![String::from("https://hogehoge.example.com/stream1"),
                               String::from("https://hogehoge.example.com/stream2"),
                               String::from("https://hogehoge.example.com/stream3"),
                               String::from("https://hogehoge.example.com/stream4"),
                               String::from("https://hogehoge.example.com/stream5"),
                               String::from("https://hogehoge.example.com/stream6"),],
            },
            Rsdata {
                station_name :  String::from("rs sample 3"),
                note : String::from("this is dummy data with two urls. https://nantoka.example.com:8080/ https://www.examples.com/nantoka"),
                tag : String::from("Talk-Sports"),
                nation : String::from("Japan"),
                _language : String::from("Japanese"),
                sources : vec![String::from("https://nantoka.exsample.com/j"),],
            },
            /* このデータがはじかれているはず
            Rsdata {
                station_name :  String::from("dummy radio 4"),
                note : String::from("this is dummy data with url. http://arienai.example.com/"),
                tag : String::from("Hard Rock"),
                nation : String::from("Japan"),
                _language : String::from("Japanese"),
                sources : vec![String::from("http://arienai.example.com/stream"),],
            },
            */
        ];

        match convert_file_to_rsdatas(&cfg, dsc) {
            Ok(rsdatas) => assert_eq!(rsdatas, org_datas),
            Err(_) => assert!(false),
        }
    }

    // -ge指定有り
    #[test]
    fn rs_convert_file_to_rsdatas_success3() {
        // config準備
        let args: Vec<String> = vec![
            "dummy".to_string(),
            "-ge=Talk,News,Public".to_string(),
            "-rs=tests/datas/radiosure_success1.rsd".to_string(),
        ];
        let cfg = match Config::new(args) {
            Ok(cfg) => cfg,
            Err(_) => {
                assert!(false);
                return;
            }
        };
        // デフォルトデータ準備(空)
        let dsc = match DefaultStationChecker::new("") {
            Ok(dsc) => dsc,
            Err(_) => {
                assert!(false);
                return;
            }
        };

        // 比較用データ準備
        let org_datas = vec![
            Rsdata {
                station_name: String::from("rs sample 1"),
                note: String::from("this is dummy data."),
                tag: String::from("Community"),
                nation: String::from("Ireland"),
                _language: String::from("English"),
                sources: vec![
                    String::from("https://tekitou.example.com/stream1"),
                    String::from("https://tekitou.example.com/stream2"),
                    String::from("https://tekitou.example.com/stream3"),
                    String::from("https://tekitou.example.com/stream4"),
                ],
            },
            Rsdata {
                station_name: String::from("rs sample 2"),
                note: String::from(
                    "this is dummy data with url. https://hogehoge.example.com:8080/",
                ),
                tag: String::from("Hard Rock"),
                nation: String::from("New Zealand"),
                _language: String::from("English"),
                sources: vec![
                    String::from("https://hogehoge.example.com/stream1"),
                    String::from("https://hogehoge.example.com/stream2"),
                    String::from("https://hogehoge.example.com/stream3"),
                    String::from("https://hogehoge.example.com/stream4"),
                    String::from("https://hogehoge.example.com/stream5"),
                    String::from("https://hogehoge.example.com/stream6"),
                ],
            },
            /* このデータがはじかれているはず
            Rsdata {
                station_name :  String::from("rs sample 3"),
                note : String::from("this is dummy data with two urls. https://nantoka.example.com:8080/ https://www.examples.com/nantoka"),
                tag : String::from("Talk-Sports"),
                nation : String::from("Japan"),
                _language : String::from("Japanese"),
                sources : vec![String::from("https://nantoka.exsample.com/j"),],
            },
            */
            Rsdata {
                station_name: String::from("dummy radio 4"),
                note: String::from("this is dummy data with url. http://arienai.example.com/"),
                tag: String::from("Hard Rock"),
                nation: String::from("Japan"),
                _language: String::from("Japanese"),
                sources: vec![String::from("http://arienai.example.com/stream")],
            },
        ];

        match convert_file_to_rsdatas(&cfg, dsc) {
            Ok(rsdatas) => assert_eq!(rsdatas, org_datas),
            Err(_) => assert!(false),
        }
    }

    // -gi指定有り
    #[test]
    fn rs_convert_file_to_rsdatas_success4() {
        // config準備
        let args: Vec<String> = vec![
            "dummy".to_string(),
            "-gi=Rock".to_string(),
            "-rs=tests/datas/radiosure_success1.rsd".to_string(),
        ];
        let cfg = match Config::new(args) {
            Ok(cfg) => cfg,
            Err(_) => {
                assert!(false);
                return;
            }
        };
        // デフォルトデータ準備(空)
        let dsc = match DefaultStationChecker::new("") {
            Ok(dsc) => dsc,
            Err(_) => {
                assert!(false);
                return;
            }
        };

        // 比較用データ準備
        let org_datas = vec![
            /* このデータははじかれているはず
            Rsdata {
                station_name :  String::from("rs sample 1"),
                note : String::from("this is dummy data."),
                tag : String::from("Community"),
                nation : String::from("Ireland"),
                _language : String::from("English"),
                sources : vec![String::from("https://tekitou.example.com/stream1"),
                               String::from("https://tekitou.example.com/stream2"),
                               String::from("https://tekitou.example.com/stream3"),
                               String::from("https://tekitou.example.com/stream4"),],
            },
            */
            Rsdata {
                station_name: String::from("rs sample 2"),
                note: String::from(
                    "this is dummy data with url. https://hogehoge.example.com:8080/",
                ),
                tag: String::from("Hard Rock"),
                nation: String::from("New Zealand"),
                _language: String::from("English"),
                sources: vec![
                    String::from("https://hogehoge.example.com/stream1"),
                    String::from("https://hogehoge.example.com/stream2"),
                    String::from("https://hogehoge.example.com/stream3"),
                    String::from("https://hogehoge.example.com/stream4"),
                    String::from("https://hogehoge.example.com/stream5"),
                    String::from("https://hogehoge.example.com/stream6"),
                ],
            },
            /* このデータがはじかれているはず
            Rsdata {
                station_name :  String::from("rs sample 3"),
                note : String::from("this is dummy data with two urls. https://nantoka.example.com:8080/ https://www.examples.com/nantoka"),
                tag : String::from("Talk-Sports"),
                nation : String::from("Japan"),
                _language : String::from("Japanese"),
                sources : vec![String::from("https://nantoka.exsample.com/j"),],
            },
            */
            Rsdata {
                station_name: String::from("dummy radio 4"),
                note: String::from("this is dummy data with url. http://arienai.example.com/"),
                tag: String::from("Hard Rock"),
                nation: String::from("Japan"),
                _language: String::from("Japanese"),
                sources: vec![String::from("http://arienai.example.com/stream")],
            },
        ];

        match convert_file_to_rsdatas(&cfg, dsc) {
            Ok(rsdatas) => assert_eq!(rsdatas, org_datas),
            Err(_) => assert!(false),
        }
    }

    // -gi & -ge指定有り
    #[test]
    fn rs_convert_file_to_rsdatas_success5() {
        // config準備
        let args: Vec<String> = vec![
            "dummy".to_string(),
            "-gi=Rock".to_string(),
            "-ge=Rock,News".to_string(),
            "-rs=tests/datas/radiosure_success1.rsd".to_string(),
        ];
        let cfg = match Config::new(args) {
            Ok(cfg) => cfg,
            Err(_) => {
                assert!(false);
                return;
            }
        };
        // デフォルトデータ準備(空)
        let dsc = match DefaultStationChecker::new("") {
            Ok(dsc) => dsc,
            Err(_) => {
                assert!(false);
                return;
            }
        };

        // 比較用データ準備
        let org_datas = vec![
            /* このデータははじかれているはず
            Rsdata {
                station_name :  String::from("rs sample 1"),
                note : String::from("this is dummy data."),
                tag : String::from("Community"),
                nation : String::from("Ireland"),
                _language : String::from("English"),
                sources : vec![String::from("https://tekitou.example.com/stream1"),
                               String::from("https://tekitou.example.com/stream2"),
                               String::from("https://tekitou.example.com/stream3"),
                               String::from("https://tekitou.example.com/stream4"),],
            },
            */
            Rsdata {
                station_name: String::from("rs sample 2"),
                note: String::from(
                    "this is dummy data with url. https://hogehoge.example.com:8080/",
                ),
                tag: String::from("Hard Rock"),
                nation: String::from("New Zealand"),
                _language: String::from("English"),
                sources: vec![
                    String::from("https://hogehoge.example.com/stream1"),
                    String::from("https://hogehoge.example.com/stream2"),
                    String::from("https://hogehoge.example.com/stream3"),
                    String::from("https://hogehoge.example.com/stream4"),
                    String::from("https://hogehoge.example.com/stream5"),
                    String::from("https://hogehoge.example.com/stream6"),
                ],
            },
            /* このデータがはじかれているはず
            Rsdata {
                station_name :  String::from("rs sample 3"),
                note : String::from("this is dummy data with two urls. https://nantoka.example.com:8080/ https://www.examples.com/nantoka"),
                tag : String::from("Talk-Sports"),
                nation : String::from("Japan"),
                _language : String::from("Japanese"),
                sources : vec![String::from("https://nantoka.exsample.com/j"),],
            },
            */
            Rsdata {
                station_name: String::from("dummy radio 4"),
                note: String::from("this is dummy data with url. http://arienai.example.com/"),
                tag: String::from("Hard Rock"),
                nation: String::from("Japan"),
                _language: String::from("Japanese"),
                sources: vec![String::from("http://arienai.example.com/stream")],
            },
        ];

        match convert_file_to_rsdatas(&cfg, dsc) {
            Ok(rsdatas) => assert_eq!(rsdatas, org_datas),
            Err(_) => assert!(false),
        }
    }

    // -ne指定有り
    #[test]
    fn rs_convert_file_to_rsdatas_success6() {
        // config準備
        let args: Vec<String> = vec![
            "dummy".to_string(),
            "-ne=New Zealand".to_string(),
            "-rs=tests/datas/radiosure_success1.rsd".to_string(),
        ];
        let cfg = match Config::new(args) {
            Ok(cfg) => cfg,
            Err(_) => {
                assert!(false);
                return;
            }
        };
        // デフォルトデータ準備(空)
        let dsc = match DefaultStationChecker::new("") {
            Ok(dsc) => dsc,
            Err(_) => {
                assert!(false);
                return;
            }
        };

        // 比較用データ準備
        let org_datas = vec![
            Rsdata {
                station_name :  String::from("rs sample 1"),
                note : String::from("this is dummy data."),
                tag : String::from("Community"),
                nation : String::from("Ireland"),
                _language : String::from("English"),
                sources : vec![String::from("https://tekitou.example.com/stream1"),
                               String::from("https://tekitou.example.com/stream2"),
                               String::from("https://tekitou.example.com/stream3"),
                               String::from("https://tekitou.example.com/stream4"),],
            },
            /* このデータははじかれているはず
            Rsdata {
                station_name :  String::from("rs sample 2"),
                note : String::from("this is dummy data with url. https://hogehoge.example.com:8080/"),
                tag : String::from("Hard Rock"),
                nation : String::from("New Zealand"),
                _language : String::from("English"),
                sources : vec![String::from("https://hogehoge.example.com/stream1"),
                               String::from("https://hogehoge.example.com/stream2"),
                               String::from("https://hogehoge.example.com/stream3"),
                               String::from("https://hogehoge.example.com/stream4"),
                               String::from("https://hogehoge.example.com/stream5"),
                               String::from("https://hogehoge.example.com/stream6"),],
            },
            */
            Rsdata {
                station_name :  String::from("rs sample 3"),
                note : String::from("this is dummy data with two urls. https://nantoka.example.com:8080/ https://www.examples.com/nantoka"),
                tag : String::from("Talk-Sports"),
                nation : String::from("Japan"),
                _language : String::from("Japanese"),
                sources : vec![String::from("https://nantoka.exsample.com/j"),],
            },
            Rsdata {
                station_name :  String::from("dummy radio 4"),
                note : String::from("this is dummy data with url. http://arienai.example.com/"),
                tag : String::from("Hard Rock"),
                nation : String::from("Japan"),
                _language : String::from("Japanese"),
                sources : vec![String::from("http://arienai.example.com/stream"),],
            },
        ];

        match convert_file_to_rsdatas(&cfg, dsc) {
            Ok(rsdatas) => assert_eq!(rsdatas, org_datas),
            Err(_) => assert!(false),
        }
    }

    // -ni指定有り
    #[test]
    fn rs_convert_file_to_rsdatas_success7() {
        // config準備
        let args: Vec<String> = vec![
            "dummy".to_string(),
            "-ni=Japan".to_string(),
            "-rs=tests/datas/radiosure_success1.rsd".to_string(),
        ];
        let cfg = match Config::new(args) {
            Ok(cfg) => cfg,
            Err(_) => {
                assert!(false);
                return;
            }
        };
        // デフォルトデータ準備(空)
        let dsc = match DefaultStationChecker::new("") {
            Ok(dsc) => dsc,
            Err(_) => {
                assert!(false);
                return;
            }
        };

        // 比較用データ準備
        let org_datas = vec![
            /* このデータははじかれているはず
            Rsdata {
                station_name :  String::from("rs sample 1"),
                note : String::from("this is dummy data."),
                tag : String::from("Community"),
                nation : String::from("Ireland"),
                _language : String::from("English"),
                sources : vec![String::from("https://tekitou.example.com/stream1"),
                               String::from("https://tekitou.example.com/stream2"),
                               String::from("https://tekitou.example.com/stream3"),
                               String::from("https://tekitou.example.com/stream4"),],
            },
            */
            /* このデータははじかれているはず
            Rsdata {
                station_name :  String::from("rs sample 2"),
                note : String::from("this is dummy data with url. https://hogehoge.example.com:8080/"),
                tag : String::from("Hard Rock"),
                nation : String::from("New Zealand"),
                _language : String::from("English"),
                sources : vec![String::from("https://hogehoge.example.com/stream1"),
                               String::from("https://hogehoge.example.com/stream2"),
                               String::from("https://hogehoge.example.com/stream3"),
                               String::from("https://hogehoge.example.com/stream4"),
                               String::from("https://hogehoge.example.com/stream5"),
                               String::from("https://hogehoge.example.com/stream6"),],
            },
            */
            Rsdata {
                station_name :  String::from("rs sample 3"),
                note : String::from("this is dummy data with two urls. https://nantoka.example.com:8080/ https://www.examples.com/nantoka"),
                tag : String::from("Talk-Sports"),
                nation : String::from("Japan"),
                _language : String::from("Japanese"),
                sources : vec![String::from("https://nantoka.exsample.com/j"),],
            },
            Rsdata {
                station_name :  String::from("dummy radio 4"),
                note : String::from("this is dummy data with url. http://arienai.example.com/"),
                tag : String::from("Hard Rock"),
                nation : String::from("Japan"),
                _language : String::from("Japanese"),
                sources : vec![String::from("http://arienai.example.com/stream"),],
            },
        ];

        match convert_file_to_rsdatas(&cfg, dsc) {
            Ok(rsdatas) => assert_eq!(rsdatas, org_datas),
            Err(_) => assert!(false),
        }
    }

    // -ni & -ne指定有り
    #[test]
    fn rs_convert_file_to_rsdatas_success8() {
        // config準備
        let args: Vec<String> = vec![
            "dummy".to_string(),
            "-ni=Japan".to_string(),
            "-ne=Japan,land".to_string(),
            "-rs=tests/datas/radiosure_success1.rsd".to_string(),
        ];
        let cfg = match Config::new(args) {
            Ok(cfg) => cfg,
            Err(_) => {
                assert!(false);
                return;
            }
        };
        // デフォルトデータ準備(空)
        let dsc = match DefaultStationChecker::new("") {
            Ok(dsc) => dsc,
            Err(_) => {
                assert!(false);
                return;
            }
        };

        // 比較用データ準備
        let org_datas = vec![
            /* このデータははじかれているはず
            Rsdata {
                station_name :  String::from("rs sample 1"),
                note : String::from("this is dummy data."),
                tag : String::from("Community"),
                nation : String::from("Ireland"),
                _language : String::from("English"),
                sources : vec![String::from("https://tekitou.example.com/stream1"),
                               String::from("https://tekitou.example.com/stream2"),
                               String::from("https://tekitou.example.com/stream3"),
                               String::from("https://tekitou.example.com/stream4"),],
            },
            */
            /* このデータははじかれているはず
            Rsdata {
                station_name :  String::from("rs sample 2"),
                note : String::from("this is dummy data with url. https://hogehoge.example.com:8080/"),
                tag : String::from("Hard Rock"),
                nation : String::from("New Zealand"),
                _language : String::from("English"),
                sources : vec![String::from("https://hogehoge.example.com/stream1"),
                               String::from("https://hogehoge.example.com/stream2"),
                               String::from("https://hogehoge.example.com/stream3"),
                               String::from("https://hogehoge.example.com/stream4"),
                               String::from("https://hogehoge.example.com/stream5"),
                               String::from("https://hogehoge.example.com/stream6"),],
            },
            */
            Rsdata {
                station_name :  String::from("rs sample 3"),
                note : String::from("this is dummy data with two urls. https://nantoka.example.com:8080/ https://www.examples.com/nantoka"),
                tag : String::from("Talk-Sports"),
                nation : String::from("Japan"),
                _language : String::from("Japanese"),
                sources : vec![String::from("https://nantoka.exsample.com/j"),],
            },
            Rsdata {
                station_name :  String::from("dummy radio 4"),
                note : String::from("this is dummy data with url. http://arienai.example.com/"),
                tag : String::from("Hard Rock"),
                nation : String::from("Japan"),
                _language : String::from("Japanese"),
                sources : vec![String::from("http://arienai.example.com/stream"),],
            },
        ];

        match convert_file_to_rsdatas(&cfg, dsc) {
            Ok(rsdatas) => assert_eq!(rsdatas, org_datas),
            Err(_) => assert!(false),
        }
    }

    // -ni & -gi指定有り
    #[test]
    fn rs_convert_file_to_rsdatas_success9() {
        // config準備
        let args: Vec<String> = vec![
            "dummy".to_string(),
            "-ni=Japan".to_string(),
            "-gi=Rock".to_string(),
            "-rs=tests/datas/radiosure_success1.rsd".to_string(),
        ];
        let cfg = match Config::new(args) {
            Ok(cfg) => cfg,
            Err(_) => {
                assert!(false);
                return;
            }
        };
        // デフォルトデータ準備(空)
        let dsc = match DefaultStationChecker::new("") {
            Ok(dsc) => dsc,
            Err(_) => {
                assert!(false);
                return;
            }
        };

        // 比較用データ準備
        let org_datas = vec![
            /* このデータははじかれているはず
            Rsdata {
                station_name :  String::from("rs sample 1"),
                note : String::from("this is dummy data."),
                tag : String::from("Community"),
                nation : String::from("Ireland"),
                _language : String::from("English"),
                sources : vec![String::from("https://tekitou.example.com/stream1"),
                               String::from("https://tekitou.example.com/stream2"),
                               String::from("https://tekitou.example.com/stream3"),
                               String::from("https://tekitou.example.com/stream4"),],
            },
            */
            /* このデータははじかれているはず
            Rsdata {
                station_name :  String::from("rs sample 2"),
                note : String::from("this is dummy data with url. https://hogehoge.example.com:8080/"),
                tag : String::from("Hard Rock"),
                nation : String::from("New Zealand"),
                _language : String::from("English"),
                sources : vec![String::from("https://hogehoge.example.com/stream1"),
                               String::from("https://hogehoge.example.com/stream2"),
                               String::from("https://hogehoge.example.com/stream3"),
                               String::from("https://hogehoge.example.com/stream4"),
                               String::from("https://hogehoge.example.com/stream5"),
                               String::from("https://hogehoge.example.com/stream6"),],
            },
            */
            /* このデータははじかれているはず
            Rsdata {
                station_name :  String::from("rs sample 3"),
                note : String::from("this is dummy data with two urls. https://nantoka.example.com:8080/ https://www.examples.com/nantoka"),
                tag : String::from("Talk-Sports"),
                nation : String::from("Japan"),
                _language : String::from("Japanese"),
                sources : vec![String::from("https://nantoka.exsample.com/j"),],
            },
            */
            Rsdata {
                station_name: String::from("dummy radio 4"),
                note: String::from("this is dummy data with url. http://arienai.example.com/"),
                tag: String::from("Hard Rock"),
                nation: String::from("Japan"),
                _language: String::from("Japanese"),
                sources: vec![String::from("http://arienai.example.com/stream")],
            },
        ];

        match convert_file_to_rsdatas(&cfg, dsc) {
            Ok(rsdatas) => assert_eq!(rsdatas, org_datas),
            Err(_) => assert!(false),
        }
    }

    // -ne & -gi指定有り
    #[test]
    fn rs_convert_file_to_rsdatas_success10() {
        // config準備
        let args: Vec<String> = vec![
            "dummy".to_string(),
            "-ne=Japan".to_string(),
            "-gi=Rock".to_string(),
            "-rs=tests/datas/radiosure_success1.rsd".to_string(),
        ];
        let cfg = match Config::new(args) {
            Ok(cfg) => cfg,
            Err(_) => {
                assert!(false);
                return;
            }
        };
        // デフォルトデータ準備(空)
        let dsc = match DefaultStationChecker::new("") {
            Ok(dsc) => dsc,
            Err(_) => {
                assert!(false);
                return;
            }
        };

        // 比較用データ準備
        let org_datas = vec![
            /* このデータははじかれているはず
            Rsdata {
                station_name :  String::from("rs sample 1"),
                note : String::from("this is dummy data."),
                tag : String::from("Community"),
                nation : String::from("Ireland"),
                _language : String::from("English"),
                sources : vec![String::from("https://tekitou.example.com/stream1"),
                               String::from("https://tekitou.example.com/stream2"),
                               String::from("https://tekitou.example.com/stream3"),
                               String::from("https://tekitou.example.com/stream4"),],
            },
            */
            Rsdata {
                station_name: String::from("rs sample 2"),
                note: String::from(
                    "this is dummy data with url. https://hogehoge.example.com:8080/",
                ),
                tag: String::from("Hard Rock"),
                nation: String::from("New Zealand"),
                _language: String::from("English"),
                sources: vec![
                    String::from("https://hogehoge.example.com/stream1"),
                    String::from("https://hogehoge.example.com/stream2"),
                    String::from("https://hogehoge.example.com/stream3"),
                    String::from("https://hogehoge.example.com/stream4"),
                    String::from("https://hogehoge.example.com/stream5"),
                    String::from("https://hogehoge.example.com/stream6"),
                ],
            },
            /* このデータははじかれているはず
            Rsdata {
                station_name :  String::from("rs sample 3"),
                note : String::from("this is dummy data with two urls. https://nantoka.example.com:8080/ https://www.examples.com/nantoka"),
                tag : String::from("Talk-Sports"),
                nation : String::from("Japan"),
                _language : String::from("Japanese"),
                sources : vec![String::from("https://nantoka.exsample.com/j"),],
            },
            */
            /* このデータははじかれているはず
            Rsdata {
                station_name :  String::from("dummy radio 4"),
                note : String::from("this is dummy data with url. http://arienai.example.com/"),
                tag : String::from("Hard Rock"),
                nation : String::from("Japan"),
                _language : String::from("Japanese"),
                sources : vec![String::from("http://arienai.example.com/stream"),],
            },
            */
        ];

        match convert_file_to_rsdatas(&cfg, dsc) {
            Ok(rsdatas) => assert_eq!(rsdatas, org_datas),
            Err(_) => assert!(false),
        }
    }

    // -ne & -ge指定有り
    #[test]
    fn rs_convert_file_to_rsdatas_success11() {
        // config準備
        let args: Vec<String> = vec![
            "dummy".to_string(),
            "-ne=Japan".to_string(),
            "-ge=Rock".to_string(),
            "-rs=tests/datas/radiosure_success1.rsd".to_string(),
        ];
        let cfg = match Config::new(args) {
            Ok(cfg) => cfg,
            Err(_) => {
                assert!(false);
                return;
            }
        };
        // デフォルトデータ準備(空)
        let dsc = match DefaultStationChecker::new("") {
            Ok(dsc) => dsc,
            Err(_) => {
                assert!(false);
                return;
            }
        };

        // 比較用データ準備
        let org_datas = vec![
            Rsdata {
                station_name: String::from("rs sample 1"),
                note: String::from("this is dummy data."),
                tag: String::from("Community"),
                nation: String::from("Ireland"),
                _language: String::from("English"),
                sources: vec![
                    String::from("https://tekitou.example.com/stream1"),
                    String::from("https://tekitou.example.com/stream2"),
                    String::from("https://tekitou.example.com/stream3"),
                    String::from("https://tekitou.example.com/stream4"),
                ],
            },
            /* このデータははじかれているはず
            Rsdata {
                station_name :  String::from("rs sample 2"),
                note : String::from("this is dummy data with url. https://hogehoge.example.com:8080/"),
                tag : String::from("Hard Rock"),
                nation : String::from("New Zealand"),
                _language : String::from("English"),
                sources : vec![String::from("https://hogehoge.example.com/stream1"),
                               String::from("https://hogehoge.example.com/stream2"),
                               String::from("https://hogehoge.example.com/stream3"),
                               String::from("https://hogehoge.example.com/stream4"),
                               String::from("https://hogehoge.example.com/stream5"),
                               String::from("https://hogehoge.example.com/stream6"),],
            },
            */
            /* このデータははじかれているはず
            Rsdata {
                station_name :  String::from("rs sample 3"),
                note : String::from("this is dummy data with two urls. https://nantoka.example.com:8080/ https://www.examples.com/nantoka"),
                tag : String::from("Talk-Sports"),
                nation : String::from("Japan"),
                _language : String::from("Japanese"),
                sources : vec![String::from("https://nantoka.exsample.com/j"),],
            },
            */
            /* このデータははじかれているはず
            Rsdata {
                station_name :  String::from("dummy radio 4"),
                note : String::from("this is dummy data with url. http://arienai.example.com/"),
                tag : String::from("Hard Rock"),
                nation : String::from("Japan"),
                _language : String::from("Japanese"),
                sources : vec![String::from("http://arienai.example.com/stream"),],
            },
            */
        ];

        match convert_file_to_rsdatas(&cfg, dsc) {
            Ok(rsdatas) => assert_eq!(rsdatas, org_datas),
            Err(_) => assert!(false),
        }
    }

    #[test]
    fn rs_convert_file_to_rsdatas_failure1() {
        // config準備
        let args: Vec<String> = vec![
            "dummy".to_string(),
            "-ge=Talk,News,Public".to_string(),
            "-ne=Russian".to_string(),
            "-rs=tests/datas/sonzai_shinai.rsd".to_string(),
        ];
        let cfg = match Config::new(args) {
            Ok(cfg) => cfg,
            Err(_) => {
                assert!(false);
                return;
            }
        };
        // デフォルトデータ準備(空)
        let dsc = match DefaultStationChecker::new("") {
            Ok(dsc) => dsc,
            Err(_) => {
                assert!(false);
                return;
            }
        };

        // データ準備
        match convert_file_to_rsdatas(&cfg, dsc) {
            Ok(_) => assert!(false),
            Err(_) => assert!(true),
        }
    }
}
