//! # defaultdatas Crate
//!
//! `defaultdatas` is library for Screamer Radio's default stations data.
//! provide one function of checking Unique station name or not.
//!
//! このCrateはScreamer Radioのデフォルトのstationデータを読み込んで
//! 指定された局名が存在するかどうかをチェックする関数を提供します。

use serde::Deserialize;
use std::collections::HashMap;
use std::error::Error;
use std::fs::File;
use std::io::BufReader;

// 以下、not pubの構造体はScreame Radioのデータファイルを読み込むための
// JSONを定義する構造体群。
// Stations.stationsとStation.title以外は一切使わない。
#[derive(Deserialize)]
struct StationLink {
    #[serde(rename = "application")]
    _application: String,
    #[serde(rename = "page")]
    _page: String,
    #[serde(rename = "website")]
    _website: String,
}

#[derive(Deserialize)]
struct StationEndlink {
    #[serde(rename = "href")]
    _href: String,
}

#[derive(Deserialize)]
struct Station {
    #[serde(rename = "href")]
    _href: String,
    #[serde(rename = "id")]
    _id: String,
    title: String,
    #[serde(rename = "links")]
    _links: StationLink,
    #[serde(rename = "tags")]
    _tags: Vec<String>,
    #[serde(rename = "endpoints")]
    _endpoints: Vec<StationEndlink>,
}

#[derive(Deserialize)]
struct DefaultStations {
    #[serde(rename = "version")]
    _version: String, // uuid
    #[serde(rename = "lastUpdate")]
    _last_update: String, // lastUpdate
    #[serde(rename = "lastFullUpdate")]
    _last_full_update: String, // lastFullUpdate
    stations: Vec<Station>,
}

impl DefaultStations {
    // This function read DefaultStations from srcname file.
    // If srcname is empty, make dummy data and return it.
    //
    // この関数はsrcnameを読み込んでDefaultStationsを返却する。
    // srcnameが空文字列の場合、ダミーデータを作成して返却する。
    fn new(srcname: &str) -> Result<DefaultStations, Box<dyn Error + 'static>> {
        if srcname.is_empty() {
            // 空の時はダミーのデータを作成して返却する
            Ok(DefaultStations {
                _version: "dummy".to_string(),
                _last_update: "dummy".to_string(),
                _last_full_update: "dummy".to_string(),
                stations: vec![],
            })
        } else {
            // ファイルを読み込む
            let file = File::open(srcname)?;
            let reader = BufReader::new(file);
            let defstations: DefaultStations = serde_json::from_reader(reader)?;
            Ok(defstations)
        }
    }
}

/// keep HashMap of Screamer Radio's default station names, and provide unique
/// name checking function.
///
/// Screamer Radioのデフォルトのstation nameをハッシュマップに保持して、指定さ
/// れたstation nameがデフォルトに無いかどうかをチェックする関数を提供します。
pub struct DefaultStationChecker {
    station_map: HashMap<String, usize>,
}

impl DefaultStationChecker {
    /// This function make DefaultStationChecker from Screamer Radio default
    /// station file.
    /// If success, return Ok(DefaultStationChecker).
    /// If failure, return Err at File::open or serde_json::from_reader.
    ///
    /// この関数はScreamer Radioのdefault station fileからDefaultStationChecker
    /// を作成します。
    /// 成功した場合はOk(DefaultStationChecker)を返却します。
    /// 失敗した場合はFile::openかserde_json::from_readerのErrを返却します。
    ///
    /// # Examples
    ///
    /// ```
    /// use rs2sr::screamerradio::defaultdatas::*;
    /// use std::process::exit;
    ///
    /// let ds_checker = match DefaultStationChecker::new("stations.json") {
    ///     Ok(chkr) => chkr,
    ///     Err(err) => {
    ///         eprintln!("failure : {}",err);
    ///         exit(1);
    ///     },
    /// };
    /// ```
    pub fn new(srcname: &str) -> Result<DefaultStationChecker, Box<dyn Error + 'static>> {
        let mut station_map = HashMap::new();

        let default_stations = DefaultStations::new(srcname)?;

        for station in default_stations.stations {
            station_map.insert(station.title, 0);
        }

        Ok(DefaultStationChecker { station_map })
    }

    /// This function check station_name what is NOT in default stations.
    /// If it is not in default, return true. Other wise return false.
    ///
    /// この関数は与えられたstation_nameがデフォルトに存在しないかをチェ
    /// ックします。もし存在しないならば真を返却します。そうで無いならば偽
    /// を返却します。
    ///
    /// # Examples
    ///
    /// ```
    /// use rs2sr::screamerradio::defaultdatas::*;
    /// use std::process::exit;
    ///
    /// let ds_checker = match DefaultStationChecker::new("stations.json") {
    ///     Ok(chkr) => chkr,
    ///     Err(err) => {
    ///         eprintln!("failure : {}",err);
    ///         exit(1);
    ///     },
    /// };
    ///
    /// if ds_checker.is_unique("dummy radio 1") {
    ///     println!("NOT in default");
    /// } else {
    ///     println!("already exists");
    /// };
    /// ```
    pub fn is_unique(&self, station_name: &str) -> bool {
        self.station_map.get(station_name).is_none()
    }
}

//
// 以下テスト
//
#[cfg(test)]
mod tests {
    use super::*;

    // 通常のdefault stations file
    #[test]
    fn dsc_success1() {
        match DefaultStationChecker::new("tests/datas/DefaultStations_success1.json") {
            Ok(dsc) => {
                assert!(!dsc.is_unique("dummy radio 1"));
                assert!(!dsc.is_unique("dummy radio 2"));
                assert!(!dsc.is_unique("dummy radio 3"));
                assert!(!dsc.is_unique("dummy radio 4"));
                assert!(!dsc.is_unique("nothing radio show"));
                assert!(dsc.is_unique("yam yam gohan"));
            }
            Err(_) => assert!(false),
        };
    }

    // 空のdefault stations file
    #[test]
    fn dsc_success2() {
        match DefaultStationChecker::new("tests/datas/DefaultStations_success2.json") {
            Ok(dsc) => {
                assert!(dsc.is_unique("dummy radio 1"));
                assert!(dsc.is_unique("dummy radio 2"));
                assert!(dsc.is_unique("dummy radio 3"));
                assert!(dsc.is_unique("dummy radio 4"));
                assert!(dsc.is_unique("nothing radio show"));
                assert!(dsc.is_unique("yam yam gohan"));
            }
            Err(_) => assert!(false),
        };
    }

    // default stations file無し
    #[test]
    fn dsc_success3() {
        match DefaultStationChecker::new("") {
            Ok(dsc) => {
                assert!(dsc.is_unique("dummy radio 1"));
                assert!(dsc.is_unique("dummy radio 2"));
                assert!(dsc.is_unique("dummy radio 3"));
                assert!(dsc.is_unique("dummy radio 4"));
                assert!(dsc.is_unique("nothing radio show"));
                assert!(dsc.is_unique("yam yam gohan"));
            }
            Err(_) => assert!(false),
        };
    }

    // 想定外のdefault stations fileその１
    #[test]
    fn dsc_failure1() {
        match DefaultStationChecker::new("tests/datas/DefaultStations_failure1.json") {
            Ok(_) => assert!(false),
            Err(err) => eprintln!("right err : {}", err),
        };
    }

    // 想定外のdefault stations fileその２
    #[test]
    fn dsc_failure2() {
        match DefaultStationChecker::new("tests/datas/DefaultStations_failure2.json") {
            Ok(_) => assert!(false),
            Err(err) => eprintln!("right err : {}", err),
        };
    }

    // 想定外のdefault stations fileその３
    #[test]
    fn dsc_failure3() {
        match DefaultStationChecker::new("tests/datas/DefaultStations_failure3.json") {
            Ok(_) => assert!(false),
            Err(err) => eprintln!("right err : {}", err),
        };
    }

    // 存在しないdefault stations file
    #[test]
    fn dsc_failure4() {
        match DefaultStationChecker::new("tests/datas/sonzai_sinai.json") {
            Ok(_) => assert!(false),
            Err(err) => eprintln!("right err : {}", err),
        };
    }
}
