//! # userdatas Crate
//!
//! `userdatas` is library for Screamer Radio's user data manipulation.
//! Only one struct, and one function provided is write to file with JSON
//! format.
//!
//! このCreateはScreamer Radioのユーザーデータを操作する為のライブラリです。
//! ユーザーデータを定義する構造体とそれをJSON形式で書き出す関数を定義します。

use serde::{Deserialize, Serialize};
use std::path::Path;
use std::fs;
use std::fs::File;

// id -> "user:"+UUID v4
// title -> title
// websiteUrl -> http://127.0.0.1/ or 存在するならばRsdate.noteから抜粋
// tags -> tag,nation
// sources -> sources
//
/// Screamer Radio's user data structure. All params are limited pub for
/// crate.
///
/// Screame Radioのユーザーデータを保持する構造体です。メンバーは全部crate限定
/// 公開となります。
#[derive(Serialize, Deserialize, Debug)]
pub struct Srdata {
    pub(crate) id: String,
    pub(crate) title: String,
    #[serde(rename = "websiteUrl")]
    pub(crate) website_url: String,
    pub(crate) tags: Vec<String>,
    pub(crate) sources: Vec<String>,
}

// Rsdataの比較(id以外をチェックする)
// debug/test用
impl Srdata {
    /// This function is for tests/debug.
    ///
    /// この関数はテスト/デバッグ用です。
    #[allow(dead_code)]
    pub fn loosy_eq(&self, r: &Srdata) -> bool {
        (self.title == r.title)
            && (self.website_url == r.website_url)
            && (self.tags == r.tags)
            && (self.sources == r.sources)
    }
}

// filenameが既に存在する場合はfilename.数字(1～100の間)に
// リネームする。
// 成功した場合はOk(_)を返す。
// 既に.100まである場合やなんらかの理由で失敗した場合はNoneを返す。
fn rename_if_needed(filename: &str) -> Option<&Path> {
    let path = Path::new(filename);
    if path.is_file() {
        // 同名ファイルが既に存在しているのでリネームしておく
        let mut done = false;
        for no in 1..100 {
            let renname = String::from(filename) + &String::from(".") + &no.to_string();
            let ren_path = Path::new(&renname);
            if !ren_path.is_file() {
                match fs::rename(path, ren_path) {
                    Ok(_) => {
                        done = true;
                        break;
                    }
                    Err(err) => {
                        eprintln!("rename fail({}) : {}", no, err);
                    }
                }
            };
        }
        if !done {
            eprintln!("rename failure....");
            return None;
        }
    }
    Some(path)
}

/// Write Screamer Radio's user datas with JSON format to dstfilename.
/// If success, return Ok(()). Otherwise failure, return Err(String).
///
/// dstfilenameにJSON formatでScreamer Radioのユーザーデータを書き込みます。
/// 書き込みに成功した場合はOk(())を、失敗した場合はErr(String)を返します。
///
/// # Examples
///
/// ```
/// use rs2sr::screamerradio::userdatas::*;
///
/// let srdatas: Vec<Srdata> = Vec::new();
/// match write_to_screamer_radio_data(&srdatas,"tests/datas/sonzai_sinai.json") {
///     Ok(_) => println!("success!"),
///     Err(msg) => eprintln!("failure : {}",msg),
/// }
/// ```
pub fn write_to_screamer_radio_data(
    srdatas: &Vec<Srdata>,
    dstfilename: &str,
) -> Result<(), String> {
    // precheck srdatas
    if srdatas.is_empty() {
        // データが空なので、なにもせずに終了。
        return Err(String::from(ERRMSG_SRDATAS_ARE_EMPTY));
    };

    //
    println!("write Screamer Radio data...");

    // ファイルの事前チェック
    let path = match rename_if_needed(dstfilename) {
        Some(path) => path,
        None => return Err(String::from(ERRMSG_CANNOT_RENAME)),
    };

    // ファイル作成
    let file = match File::create(path) {
        Ok(f) => f,
        Err(err) => return Err(format!("{}{}", ERRMSG_FILE_CREATE_ERR, err)),
    };

    // ファイル書き込み
    if let Err(err) = serde_json::to_writer_pretty(file, srdatas) {
        // 書き込み失敗
        return Err(format!("{}{}", ERRMSG_FILE_WRITE_ERR, err));
    }

    // 完了
    Ok(())
}
const ERRMSG_SRDATAS_ARE_EMPTY: &str = "Datas is empty. Do nothing !";
const ERRMSG_CANNOT_RENAME: &str = "cannot rename file";
const ERRMSG_FILE_CREATE_ERR: &str = "file create err : ";
const ERRMSG_FILE_WRITE_ERR: &str = "file write err : ";

//
// 以下テスト
//
#[cfg(test)]
mod tests {
    use super::*;
    use std::io::{BufReader, ErrorKind};

    #[test]
    // リネームの必要が無いパターン
    fn srs_rename_if_needed_success1() {
        let name: &str = "tests/datas/rn_work/sonzai_sinai.txt";
        let chk_path = Path::new(name);

        match rename_if_needed(name) {
            Some(path) => assert_eq!(path, chk_path),
            None => assert!(false),
        }
    }

    #[test]
    // リネームの必要が有るパターン
    fn srs_rename_if_needed_success2() {
        let name: &str = "tests/datas/rn_work/hitotu_aru_pattern.txt";
        let r_name: &str = "tests/datas/rn_work/hitotu_aru_pattern.txt.1";

        // rename先を削除しておく
        match fs::remove_file(r_name) {
            Ok(_) => {}
            Err(err) => assert_eq!(err.kind(), ErrorKind::NotFound),
        }

        // 必要ならば同名ファイルを作成しておく
        let path = Path::new(name);
        if !path.is_file() {
            match File::create(path) {
                Ok(_) => {}
                Err(_) => assert!(false),
            };
        }

        // rename実行
        assert_ne!(rename_if_needed(name), None);

        // 存在チェック
        assert!(!path.is_file()); // ファイルが存在してはいけない
        let r_path = Path::new(r_name);
        assert!(r_path.is_file()); // ファイルが存在していなければならない
    }

    #[test]
    // リネーム失敗
    fn srs_rename_if_needed_failure() {
        let name: &str = "tests/datas/rn_work/sudeni_aru.json";

        assert_eq!(rename_if_needed(name), None); // 失敗している必要がある
    }

    #[test]
    // データ書き込み成功
    fn srs_write_to_screamer_radio_data_success() {
        // データの用意
        let org_datas = vec![
            Srdata {
                id: "user:uuid_v4_tekitou01".to_string(),
                title: "hogehoge".to_string(),
                website_url: "https://stream.example.com/".to_string(),
                tags: vec!["User".to_string()],
                sources: vec![
                    "https://www.example.com/jazz".to_string(),
                    "https://www.example.com:8080/jazz".to_string(),
                ],
            },
            Srdata {
                id: "user:uuid_v4_tekitou02".to_string(),
                title: "fugafuga".to_string(),
                website_url: "https://boke.example.com/;".to_string(),
                tags: vec![
                    "Pop".to_string(),
                    "New Zealand".to_string(),
                    "User".to_string(),
                ],
                sources: vec![
                    "https://boke.example.com/;".to_string(),
                    "https://boke.example.com:8080/pop".to_string(),
                ],
            },
            Srdata {
                id: "user:uuid_v4_tekitou03".to_string(),
                title: "mocho mocho".to_string(),
                website_url: "https://mochimochi.example.com:8080/?stat=neko".to_string(),
                tags: vec!["Jazz".to_string(), "User".to_string()],
                sources: vec!["https://mochi.example.com:8080/".to_string()],
            },
        ];
        let srdatas_name = "tests/datas/test_mystation.json";

        // 事前準備
        match fs::remove_file(srdatas_name) {
            Ok(_) => {}
            Err(err) => assert_eq!(err.kind(), ErrorKind::NotFound),
        }

        // 書き込み
        match write_to_screamer_radio_data(&org_datas, srdatas_name) {
            Ok(_) => {
                // 存在確認
                let path = Path::new(srdatas_name);
                assert!(path.is_file());
                // 読み込み
                let file = match File::open(&path) {
                    Err(err) => {
                        assert!(false, "{}", err);
                        return;
                    }
                    Ok(file) => file,
                };
                let reader = BufReader::new(file);
                // 中身の確認
                match serde_json::from_reader::<BufReader<File>, Vec<Srdata>>(reader) {
                    Ok(datas) => {
                        assert!(datas.len() == org_datas.len(), "differnt data lens.");
                        let mut r_iter = datas.iter();
                        for l in org_datas {
                            let r = r_iter.next().unwrap();
                            assert!(l.loosy_eq(&r));
                        }
                    }
                    Err(err) => assert!(false, "{}", err),
                }
            }
            Err(msg) => assert!(false, "{}", msg),
        }
    }

    #[test]
    // データが空
    fn srs_write_to_screamer_radio_data_failure1() {
        let empty_data: Vec<Srdata> = vec![];

        match write_to_screamer_radio_data(&empty_data, "tests/datas/tekitou.json") {
            Ok(_) => assert!(false),
            Err(err) => assert_eq!(err, ERRMSG_SRDATAS_ARE_EMPTY),
        }
    }

    #[test]
    // リネームできない
    fn srs_write_to_screamer_radio_data_failure2() {
        let datas: Vec<Srdata> = vec![Srdata {
            id: "user:uuid_v4_tekitou01".to_string(),
            title: "hogehoge".to_string(),
            website_url: "https://stream.example.com/".to_string(),
            tags: vec!["User".to_string()],
            sources: vec![
                "https://www.example.com/jazz".to_string(),
                "https://www.example.com:8080/jazz".to_string(),
            ],
        }];

        match write_to_screamer_radio_data(&datas, "tests/datas/rn_work/sudeni_aru.json") {
            Ok(_) => assert!(false),
            Err(err) => assert_eq!(err, ERRMSG_CANNOT_RENAME),
        }
    }

    #[test]
    // ファイル作成できない
    fn srs_write_to_screamer_radio_data_failure3() {
        let datas: Vec<Srdata> = vec![Srdata {
            id: "user:uuid_v4_tekitou01".to_string(),
            title: "hogehoge".to_string(),
            website_url: "https://stream.example.com/".to_string(),
            tags: vec!["User".to_string()],
            sources: vec![
                "https://www.example.com/jazz".to_string(),
                "https://www.example.com:8080/jazz".to_string(),
            ],
        }];

        match write_to_screamer_radio_data(&datas, "tests/datas/dummy.json") {
            Ok(_) => assert!(false),
            Err(err) => assert!(err.contains(ERRMSG_FILE_CREATE_ERR), "err={}", err),
        }
    }

    #[test]
    // ファイル書き込み失敗 → テストが思いつかないので今のところ無し
    fn srs_write_to_screamer_radio_data_failure4() {}
}
